package eu.jonahbauer.wizard.client.libgdx.screens;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.Interpolation;
import com.badlogic.gdx.scenes.scene2d.Actor;
import com.badlogic.gdx.scenes.scene2d.ui.*;
import com.badlogic.gdx.scenes.scene2d.ui.Stack;
import com.badlogic.gdx.scenes.scene2d.utils.ChangeListener;
import com.badlogic.gdx.scenes.scene2d.utils.TextureRegionDrawable;
import com.badlogic.gdx.utils.Align;
import eu.jonahbauer.wizard.client.libgdx.GameAtlas;
import eu.jonahbauer.wizard.client.libgdx.UiskinAtlas;
import eu.jonahbauer.wizard.client.libgdx.WizardGame;
import eu.jonahbauer.wizard.client.libgdx.actors.game.*;
import eu.jonahbauer.wizard.client.libgdx.util.Pair;
import eu.jonahbauer.wizard.common.messages.observer.*;
import eu.jonahbauer.wizard.common.messages.server.GameMessage;
import eu.jonahbauer.wizard.common.messages.server.ServerMessage;
import eu.jonahbauer.wizard.common.model.Card;
import lombok.Getter;

import java.util.*;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.*;

public class GameScreen extends MenuScreen {
    private final WizardGame game;

    private TextureAtlas atlas;
    private TextureRegion background;

    private final List<UUID> players;

    private final UUID self;
    private final UUID session;
    private final Map<UUID, String> names;
    private final Map<Integer, Map<UUID, Integer>> scores = new HashMap<>();

    private int round = -1;
    private final Map<UUID, List<Card>> hands = new HashMap<>();
    private final Map<UUID, Integer> predictions = new HashMap<>();
    private final Map<UUID, List<List<Card>>> tricks = new HashMap<>();
    private Card.Suit trumpSuit;
    private Card trumpCard;

    private int trick = -1;
    private final List<Pair<UUID, Card>> stack = new ArrayList<>();
    private Pair<UUID, UserInputMessage.Action> activePlayer;

    private CardsGroup handCards;
    private CardStack cardStack;
    private PadOfTruth padOfTruth;
    private CardActor trumpCardActor;
    private Map<UUID, Seat> seats = new HashMap<>();
    private Map<UUID, Label> nameLabels = new HashMap<>();

    public GameScreen(WizardGame game) {
        this(game, null);
    }

    public GameScreen(WizardGame game, Data data) {
        super(game, data);
        this.game = game;

        //<editor-fold desc="Sample Data" defaultstate="collapsed">
        this.session = UUID.randomUUID();
        this.names = new HashMap<>();

        int playerCount = 6;
        var players = new UUID[playerCount];
        for (int i = 0; i < playerCount; i++) {
            players[i] = UUID.randomUUID();
            names.put(players[i], "Player " + i);
        }

        this.self = players[0];
        this.players = Stream.of(players).sorted().toList();
        //</editor-fold>
    }

    @Override
    public void show() {
        super.show();
        atlas = new TextureAtlas(Gdx.files.internal(GameAtlas.$PATH));
        background = atlas.findRegion(GameAtlas.BACKGROUND);

        seat();
        prepareLabels();

        handCards = new CardsGroup(Collections.emptyList(), atlas);
        handCards.setOnClickListener(card -> playCard(self, card.getCard()));
        var container = new Container<>(handCards);
        container.setPosition(360, 75);
        container.setSize(1200, 210);
        container.fillX();

        padOfTruth = new PadOfTruth(data.skin, new TextureRegionDrawable(atlas.findRegion(GameAtlas.PAD_OF_TRUTH)));
        padOfTruth.setPosition(1910 - 636f, 10);
        padOfTruth.setOrigin(636f, 0);

        cardStack = new CardStack();
        cardStack.setHoverBounds(0.5f * WizardGame.WIDTH - 50, 0.5f * WizardGame.HEIGHT - 50, 100, 100);

        setNames();

        var trump = new TextButton("Trump", data.skin);
        trump.align(Align.left);
        trump.addListener(new ChangeListener() {
            @Override
            public void changed(ChangeEvent event, Actor actor) {
                setTrumpCard(Card.values()[(int)(Math.random() * Card.values().length)]);
            }
        });

        var playCard = new TextButton("Play Card", data.skin);
        playCard.addListener(new ChangeListener() {
            @Override
            public void changed(ChangeEvent event, Actor actor) {
                var hand = hands.get(self);
                if (hand != null && hand.size() > 0) {
                    playCard(self, hand.get((int) (Math.random() * hand.size())));
                }
            }
        });

        var deal = new TextButton("Deal", data.skin);
        deal.addListener(new ChangeListener() {
            @Override
            public void changed(ChangeEvent event, Actor actor) {
                var hand = Stream.of(Card.values()).limit(20).toList();
                setHand(self, hand);
            }
        });

        var round = new TextButton("Round", data.skin);
        round.addListener(new ChangeListener() {
            @Override
            public void changed(ChangeEvent event, Actor actor) {
                startRound();
                round.setText("Round " + GameScreen.this.round);
            }
        });

        var scores = new TextButton("Scores", data.skin);
        scores.addListener(new ChangeListener() {
            @Override
            public void changed(ChangeEvent event, Actor actor) {
                var scores = players.stream().collect(Collectors.toMap(p -> p, p -> (int) (Math.random() * 5) * 10));
                addScores(scores);
            }
        });

        var predictions = new TextButton("Predictions", data.skin);
        predictions.addListener(new ChangeListener() {
            @Override
            public void changed(ChangeEvent event, Actor actor) {
                var predictions = players.stream().collect(Collectors.toMap(p -> p, p -> (int) (Math.random() * 5)));
                predictions.forEach((player, prediction) -> addPrediction(player, prediction));
            }
        });

        var playOtherCard = new TextButton("Play Other Card", data.skin);
        playOtherCard.addListener(new ChangeListener() {
            @Override
            public void changed(ChangeEvent event, Actor actor) {
                var card = Card.values()[(int) (Math.random() * Card.values().length)];
                UUID player;
                do {
                    player = players.get((int) (Math.random() * players.size()));
                } while (player.equals(self));
                playCard(player, card);
            }
        });

        var buttons = new VerticalGroup();
        buttons.top().left().columnLeft().setPosition(0, WizardGame.HEIGHT - buttons.getHeight());
        buttons.addActor(round);
        buttons.addActor(deal);
        buttons.addActor(trump);
        buttons.addActor(predictions);
        buttons.addActor(playCard);
        buttons.addActor(playOtherCard);
        buttons.addActor(scores);

        Gdx.input.setInputProcessor(data.stage);
        data.stage.addActor(container);
        data.stage.addActor(cardStack);
        data.stage.addActor(padOfTruth);
        data.stage.addActor(buttons);
    }

    @Override
    protected void renderBackground(float delta) {
        float scale = Math.max(
                data.extendViewport.getWorldWidth() / WizardGame.WIDTH,
                data.extendViewport.getWorldHeight() / WizardGame.HEIGHT
        );
        game.batch.draw(data.background, 0,0, scale * WizardGame.WIDTH, scale * WizardGame.HEIGHT);
        game.batch.setColor(1, 1, 1, 0.25f);
        game.batch.draw(
                data.title,
                (data.extendViewport.getWorldWidth() - data.title.getRegionWidth() * 0.75f) / 2f,
                (data.extendViewport.getWorldHeight() - data.title.getRegionHeight() * 0.75f) / 2f,
                0.75f * data.title.getRegionWidth(),
                0.75f * data.title.getRegionHeight()
        );
    }

    @Override
    protected void renderForeground(float delta) {}

    private void seat() {
        var count = players.size();
        Seat[] seats = switch (count) {
            case 3 -> new Seat[] {Seat.TOP_LEFT, Seat.TOP_RIGHT};
            case 4 -> new Seat[] {Seat.LEFT, Seat.TOP, Seat.RIGHT};
            case 5 -> new Seat[] {Seat.LEFT, Seat.TOP_LEFT, Seat.TOP_RIGHT, Seat.RIGHT};
            case 6 -> new Seat[] {Seat.LEFT, Seat.TOP_LEFT, Seat.TOP, Seat.TOP_RIGHT, Seat.RIGHT};
            default -> throw new AssertionError();
        };
        int index = players.indexOf(self);
        for (int i = 1; i < count; i++) {
            var player = players.get((index + i) % count);
            var seat = seats[i - 1];
            this.seats.put(player, seat);
        }
    }

    private void prepareLabels() {
        for (UUID player : players) {
            if (player.equals(self)) continue;
            var label = new Label("", data.skin);
            var seat = seats.get(player);
            label.setX(seat.getLabelX());
            label.setY(seat.getLabelY());
            label.setAlignment(seat.getLabelAlign());
            this.nameLabels.put(player, label);
            data.stage.addActor(label);
        }
    }

    private void setNames() {
        for (int i = 0; i < players.size(); i++) {
            var player = players.get(i);
            var name = names.get(player);
            padOfTruth.setName(i, name);
            if (!self.equals(player)) {
                nameLabels.get(player).setText(name);
            }
        }
    }

    private void startRound() {
        round++;
        hands.clear();
        tricks.clear();
        predictions.clear();
        trumpSuit = null;
        trumpCard = null;
        handCards.clearChildren();
        cardStack.clearChildren();
        if (trumpCardActor != null) {
            trumpCardActor.remove();
        }
    }

    private void setHand(UUID player, List<Card> cards) {
        var hand = hands.computeIfAbsent(player, p -> new ArrayList<>());
        hand.clear();
        hand.addAll(cards);

        handCards.update(hand);
    }

    private void setTrumpCard(Card trumpCard) {
        this.trumpCard = trumpCard;
        if (trumpCardActor != null) {
            trumpCardActor.remove();
        }

        if (trumpCard == null) return;

        if (trumpCardActor != null) {
            trumpCardActor.setCard(trumpCard);
        } else {
            trumpCardActor = new CardActor(trumpCard, atlas);
        }

        var background = new Container<>();
        background.setBackground(new TextureRegionDrawable(data.uiskinAtlas.findRegion(UiskinAtlas.WHITE)));
        background.setColor(0.0f, 0.0f, 0.0f, 0.5f);
        background.setFillParent(true);

        var screen = new Stack();
        screen.setFillParent(true);
        screen.add(background);

        var group = new VerticalGroup().align(Align.center);
        screen.add(group);

        var label = new Label("Determining trump", data.skin);
        label.setFontScale(1.5f);
        group.addActor(label);
        group.addActor(trumpCardActor);

        data.stage.addAction(sequence(
                run(() -> data.stage.addActor(screen)),
                delay(5.0f),
                removeActor(label),
                addAction(moveTo(10,10, .25f), trumpCardActor),
                delay(.25f),
                addAction(alpha(0.0f, .1f, Interpolation.pow2Out), background),
                delay(.1f),
                run(() -> data.stage.addActor(trumpCardActor)),
                removeActor(screen)
        ));
    }

    private void addPrediction(UUID player, int prediction) {
        predictions.put(player, prediction);
        padOfTruth.setPrediction(players.indexOf(player), round, prediction);
    }

    private void playCard(UUID player, Card card) {
        var handCard = switch (card) {
            case CHANGELING_JESTER, CHANGELING_WIZARD -> Card.CHANGELING;
            case CLOUD_BLUE, CLOUD_RED, CLOUD_GREEN, CLOUD_YELLOW -> Card.CLOUD;
            case JUGGLER_BLUE, JUGGLER_RED, JUGGLER_GREEN, JUGGLER_YELLOW -> Card.JUGGLER;
            default -> card;
        };

        if (hands.containsKey(player)) {
            hands.get(player).remove(handCard);
        }
        stack.add(Pair.of(player, card));

        CardActor actor;
        Seat seat;
        if (player.equals(self)) {
            actor = handCards.remove(handCard);
            seat = Seat.BOTTOM;
        } else {
            actor = new CardActor(card, atlas);
            seat = seats.get(player);
            actor.setPosition(seat.getX() - actor.getWidth() / 2, seat.getY() - actor.getHeight() / 2);
            actor.setRotation(seat.getAngle());
        }
        actor.setOrigin(actor.getWidth() / 2, actor.getHeight() / 2);
        cardStack.add(seat, actor);
    }

    private void addScores(Map<UUID, Integer> scores) {
        this.scores.put(round, scores);

        for (int i = 0; i < players.size(); i++) {
            UUID player = players.get(i);
            padOfTruth.setScore(i, round, scores.get(player));
        }
    }

    @Override
    public void dispose() {
        super.dispose();
        if (atlas != null) atlas.dispose();
    }

    public void onMessage(ServerMessage serverMessage) {
        if (serverMessage instanceof GameMessage gameMessage) {
            var observerMessage = gameMessage.getObserverMessage();
            if (observerMessage instanceof CardMessage card) {
                stack.add(Pair.of(card.getPlayer(), card.getCard()));
            } else if (observerMessage instanceof TrickMessage trick) {
                stack.clear();
                tricks.computeIfAbsent(trick.getPlayer(), p -> new ArrayList<>()).add(trick.getCards());
            } else if (observerMessage instanceof HandMessage hand) {
                var cards = hands.computeIfAbsent(hand.getPlayer(), p -> new ArrayList<>());
                cards.clear();
                cards.addAll(hand.getHand());
            } else if (observerMessage instanceof ScoreMessage score) {
                var roundScores = scores.computeIfAbsent(round, r -> new HashMap<>());
                roundScores.putAll(score.getPoints());
            } else if (observerMessage instanceof TrumpMessage trump) {
                trumpCard = trump.getCard();
                trumpSuit = trump.getSuit();
                //Todo update TrumpCardActor if trumpSuit != null and trumpCard != null
            } else if (observerMessage instanceof PredictionMessage prediction) {
                predictions.put(prediction.getPlayer(), prediction.getPrediction());
                //TODO update table and label
            } else if (observerMessage instanceof UserInputMessage userInput) {
                activePlayer = Pair.of(userInput.getPlayer(), userInput.getAction());
                if (self.equals(userInput.getPlayer())) {
                    // TODO do something
                }
            } else if (observerMessage instanceof StateMessage state) {
                switch (state.getState()) {
                    case "starting_round" -> {
                        round++;
                        trick = -1;
                        hands.clear();
                        predictions.clear();
                        tricks.clear();
                        trumpSuit = null;
                        trumpCard = null;
                    }
                    case "starting_trick" -> {
                        trick++;
                        stack.clear();
                        activePlayer = null;
                    }
                    case "finished", "error" -> {
                        // TODO do something
                    }
                }
            }
        }
    }

    @Getter
    public enum Seat {
        BOTTOM(WizardGame.WIDTH * 0.5f, 0, 0, 0, Align.bottom, WizardGame.WIDTH * 0.5f, 300),
        LEFT(0, WizardGame.HEIGHT * 0.5f, 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomLeft, 117.5f, WizardGame.HEIGHT * 0.5f),
        RIGHT(WizardGame.WIDTH, WizardGame.HEIGHT * 0.5f, WizardGame.WIDTH - 50, WizardGame.HEIGHT * 0.5f + 110f, Align.bottomRight, WizardGame.WIDTH - 117.5f, WizardGame.HEIGHT * 0.5f),
        TOP_LEFT(WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.25f, WizardGame.HEIGHT - 200),
        TOP(WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.5f, WizardGame.HEIGHT - 200),
        TOP_RIGHT(WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 50, Align.top, WizardGame.WIDTH * 0.75f, WizardGame.HEIGHT - 200);

        private final float x;
        private final float y;
        private final float labelX;
        private final float labelY;
        private final int labelAlign;
        private final float angle;
        private final float cardX;
        private final float cardY;

        Seat(float x, float y, float labelX, float labelY, int labelAlign, float cardX, float cardY) {
            this.x = x;
            this.y = y;
            this.labelX = labelX;
            this.labelY = labelY;
            this.labelAlign = labelAlign;
            this.cardX = cardX;
            this.cardY = cardY;

            var deltaX = WizardGame.WIDTH * 0.5f - x;
            var deltaY = WizardGame.HEIGHT * 0.5f - y;
            this.angle = (float) Math.toDegrees(Math.atan2(deltaY, deltaX) + Math.PI / 2);
        }
    }
}
